#!/usr/bin/perl -w
###############################################################################
#
#       PROGRAM NAME- omwe2sval.pl
#	This program converts the sense-tagged data created by the Open-Mind
#	project to senseval2 format.

###############################################################################
#                       ---------------------
#                       COPYRIGHT INFORMATION
#                       ---------------------

# Copyright (c) 2002,
# Amruta Purandare, University of Minnesota, Duluth
# pura0010@umn.edu
# Ted Pedersen, University of Minnesota, Duluth
# tpederse@umn.edu
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to
#
# The Free Software Foundation, Inc.,
# 59 Temple Place - Suite 330,
# Boston, MA  02111-1307, USA.
###############################################################################

#                       =============================
#                            THE CODE STARTS HERE
#                       =============================

##############################################################################
#                       ===============================
#                       COMMAND LINE OPTIONS AND USAGE
#                       ===============================

#show minimal usage
if($#ARGV<0)
{
        &minimal();
        exit;
}

use Getopt::Long;
GetOptions ("help","version");

#command option for help
if(defined $opt_help)
{
        $opt_help=1;
        &showhelp();
        exit;
}
#version information
if(defined $opt_version)
{
        $opt_version=1;
        &showversion();
        exit;
}

##############################################################################

#                       ================================
#                          INITIALIZATION AND INPUT
#                       ================================

if(!defined $ARGV[0])
{
	print STDERR "ERROR:
        Please specify the name of the file that maps the instance-ids to the
	sense tags assigned by the contributors...\n";
        exit;
}
if(!defined $ARGV[1])
{
        print STDERR "ERROR:
Please specify the name of the file that maps the instance-ids to the 
corresponding instances...\n";
        exit;
}

#accept the name of the file that contains the ids to sense tag mapping
$id2tag=$ARGV[0];
$id2instance=$ARGV[1];

#check if the files exist and open if they do
if(!(-e $id2tag))
{
	print STDERR "ERROR:
The file mapping the instance id to the sense tags $id2tag doesn't exist...\n";
        exit;
}

if(!(-e $id2instance))
{
        print STDERR "ERROR:
The file mapping the instance id to the instances $id2instance doesn't exist...\n";
        exit;
}

open(ID2TAG,$id2tag) || die "Error in opening $id2tag file.\n";
open(ID2INSTANCE,"$id2instance") || die "Error in opening $id2instance file.\n";

##############################################################################

#			==========================
#			  Actual Conversion Logic			
#			==========================

# Reading a map file that maps the instance ids to the sense tags 
# we build a table here that can be referred to get the tags associated 
# for an instance by several contributors 
while($line=<ID2TAG>)
{
	chomp $line;
	$line=~s/^\s+//g;
	$line=~s/\s$//g;
	$line=~s/\s+/ /g;
	@entry=split(/\s+/,$line);
	#blank lines in file
	if($#entry<0)
	{
		next;
	}
	# case when more than one tags is assigned by the user for the same 
	# sense which is an invalid case according to the open-mind README 
	# we dont stop processing neither we skip this line but just show a 
	# warning message 
	if($#entry>1)
	{
		print STDERR "Warning:
	The entry $line 
	has more than one tags by the same user for the same instance...\n";
	}
	# get the instance id part
	if(defined $entry[0])
	{
		$current_id=$entry[0];
	}
	else
	{
		next;
	}
	# generally only 1 tag but it is possible in exceptional cases to have
	# more than one tags per contributor per item
	foreach (1..$#entry)
	{
		$map_tab{$current_id}.="$entry[$_]\t";
	}
	if($#entry==0)	
	{
		$map_tab{$current_id}.="blank\t";
	}
}

###	finished the part of building the map table here ###
###	now the format conversion logic !!!		 ###

# open various output files 
# notag file lists all the instances for which it doesn't find any tag in the 
# tag file
open(OP1_FILE,">output/notag.txt") || die "Error in opening output file notag.txt\n";

# mismatch file lists all the instance ids for which the double check of the 
# target word occuring at the specified target location fails
open(OP2_FILE,">output/mismatch.txt") || die "Error in opening output file mismatch.txt\n";

# repeated.txt file lists all the instances which occur multiple times in the 
# instance file
open(OP3_FILE,">output/repeated.txt") || die "Error in opening output file repeated.txt\n";

# instances in the tag file with no entry in the instance file are written to
# the nointance.txt file
open(OP4_FILE,">output/noinstance.txt") || die "Error in opening output file noinstance.txt\n";

# typical senseval2 data format starts with these standard tags
print "<?xml version=\"1.0\" encoding=\"iso-8859-1\"?>\n";
print "<!DOCTYPE corpus SYSTEM \"lexical-sample.dtd\">\n";
print "<corpus lang='english'>\n";
# reading the file containing the sense tagged data along with the
# instance ids
$flag=0;
while($line=<ID2INSTANCE>)
{
	# traditional trimming operations to get rid of the extra spaces
	chomp $line;
	$line=~s/^\s+//g;
        $line=~s/\s$//g;
	# parts of the instances are space separated 
        @sentence=split(/\s+/,$line);
	if($#sentence<0)	
	{
		next;	
	}
	# first part in an instance must be the instance id
	if(!defined $sentence[0])
	{
		print STDERR "Warning:
	Couldn't find instance-id at line <$line>\n";
		next;
	}
	$instance_id=$sentence[0];	
	# checking for the multiple entries in the file id2instance
	if(!defined $multi{$instance_id})
	{
		$multi{$instance_id}=1;
	}
	else
	{
		$multi{$instance_id}++;
		next;
	}
	# second part of the instance states the taregt word
	if(!defined $sentence[1])
        {
                print STDERR "Warning:
        Couldn't find target word at line <$line>\n";
        	next;
        }
	$target=$sentence[1];
	if(!defined $sentence[3])
        {
                print STDERR "Warning:
        Couldn't find target word location at line <$line>\n";
        	next;
        }
	# third part gives the target word position in the current instance
	$target_location=$sentence[3];
	# target word in its base form is found in the beginning of the 
	# instance id which is . separated
	($part1,$part2,@rest)=split(/\./,$instance_id);
	undef @rest;
	if((!defined $part1) || (!defined $part2))
        {
                print STDERR "ERROR:
        The instance id format at <$line> is invalid.\n";
                exit;
        }
	$word="$part1."."$part2";
	# get all the senses for this instance assigned by different
        # contributors from the map table
        if(defined $map_tab{$instance_id})
        {
                @senses=split(/\s+/,$map_tab{$instance_id});
        }
        else
        {
                print OP1_FILE "$instance_id\n";
                next;
        }
	# new lexelt tag when the new target word occurs
	if(!defined $pre_word)
	{
		$pre_word=$word;
		print "<lexelt item=\"$word\">\n";
		$flag=1;
	}
	else
	{
		if($pre_word ne $word)
		{
			print "</lexelt>\n<lexelt item=\"$word\">\n";
			$pre_word=$word;
		}
	}
	# each instance begins with this tag which gives the instance id
        # in its attribute list
        print "<instance id=\"$instance_id\">\n";
	# each answer tag here lists a possible sense for the instance
	foreach $sense (@senses)
	{
		print "<answer instance=\"$instance_id\" senseid=\"$sense\"/>\n";
	}
	# the acutal tokens in an instance follow the context tag
	print "<context>\n";
	foreach $token_pos (4..$#sentence)	
	{
		# extract the pos and the named entity information 
		# for the token which is / separated 
		($token,@extras)=split(/\//,$sentence[$token_pos]);
		# insert head tags around the target word
		if($token_pos == ($target_location+4))			
		{
			# double checking when the target word specified 
			# in the beginning of the instance doesn't match 
			# with the actual token found at the target location
			if($token ne $target)
			{
				#double check fails !!!
				print OP2_FILE "$instance_id\n";
			}
			print " <head>";
		}
		# token is followed by the pos tag
		if($#extras>=0)
		{
			print " $token <p=\"$extras[0]\"/>";
		}
		# if present insert the named entity tag as well
		if($#extras==1)
		{
			print " <ne=\"$extras[1]\"/>";
		}	
		# if the open mind data also lists anything other than the 
		# pos and the named entity information
		if($#extras>1)
		{
			print STDERR "Warning:
	Found more than 2 extra tags in <$line>\n";
		}
		# end of the head tag
		if($token_pos == ($target_location+4))
                {
                        print " </head>";
                }

	}
	# context ends here
	print "\n</context>\n";
	# instance ends here
	print "</instance>\n";
}
# writing the repeated instances in repeated.txt file along with the 
# counts showing how many times each instance occurs 
foreach $instance (sort keys %multi)
{
	if($multi{$instance}>1)
	{
		print OP3_FILE "$instance\t\t$multi{$instance}\n";
	}
}
if($flag==1)
{
	# final lexelt tag
	print "</lexelt>\n";
}
# the corpus ends here
print "</corpus>\n";
# writing the instances found in tag file 
# but not in the instance file to the noinstance.err  
foreach $inst (keys %map_tab)
{
	if(!defined $multi{$inst})	
	{
		print OP4_FILE "$inst\n";
	}
}
##############################################################################

#                       ==========================
#                           SUBROUTINE SECTION 
#                       ==========================

#-----------------------------------------------------------------------------
#show minimal usage message
sub minimal()
{
        print "Usage: omwe2sval.pl [OPTIONS] TAG_FILE INSTANCE_FILE";
        print "\nTYPE omwe2sval.pl --help for help\n";
}
#-----------------------------------------------------------------------------#show help
sub showhelp()
{
        print "This program converts the sense tagged data collected by the Open Mind Word 
Experts into SENSEVAL2 format.";

        print "
Usage: omwe2sval.pl [OPTIONS] TAG_FILE INSTANCE_FILE\n";
        print "
TAG_FILE
	Each line of the TAG_FILE should show a tag assigned to an instance
	in the following format 
        I       TAG
        Where I is an instance id and TAG shows a sense tag assigned to I
        by a contributor.\n";
	print "
INSTANCE_FILE
	Specify the name of the file that contains the Instance data along with
	an identifier for each instance. The INSTANCE_FILE should follow the 
	format specified by the Open-Mind README that expects each instance on
	a separate line in the following format  
	I target_word ? target_location Word/POS[/NE] [Word/POS[/NE] ..] 
	where 	
	I 		= Instance Id
	target_word 	= Target word as it appears in instance I
	target_location = Location at which the target_word is found
			  in instance I when the words are counted from
			  0. 	
	Word/POS[/NE]	= Word in instance I with POS tag 
			  and optional Named Entity Information 
			  separated by /
	
	TAG_FILE and INSTANCE_FILE should follow the format specified by the
	Open Mind README. omwe2sval.pl will terminate with error if the 
	instance-ids found in these files do not follow the standard format.\n";
        print "
OPTIONS: 
--help
        Get the help message.";
        print "
--version
        Get the version information.\n";
}

#------------------------------------------------------------------------------
#version information
sub showversion()
{
        print "omwe2sval.pl      -       Version 0.01";
	print "
The program that converts the Open-Mind data into Senseval2 format.\n";
        print "Copyright (C) 2002, Amruta Purandare and Ted Pedersen.\n";
        print "Date of Last Update:     12/14/2002\n";

}

